/*******************************************************************************
 * CS559 Spring 2008 - Project 2 - Trains and Roller Coasters
 *
 * Example Solution, by Chi Man Liu
 * begin 3/13/2008
 *
 * File: P2GlWindow.cpp
 * Implements the P2GlWindow class. This file is long - about 50% dedicated to
 * drawing.
 ******************************************************************************/
#include "P2GlWindow.h"

// the FlTk headers have lots of warnings - these are bad, but there's not
// much we can do about them. We enclose the #include's by #pragma's to
// suppress warnings generated by the Fltk library.
#pragma warning(push)
#pragma warning(disable:4311)		// convert void* to long
#pragma warning(disable:4312)		// convert long to void*
#include <Fl/Fl.h>
#include <Fl/Gl.h>
#include <Gl/Glu.h>
#pragma warning(pop)

// constructor
P2GlWindow::P2GlWindow(int x, int y, int width, int height, char* title)
				: Fl_Gl_Window(x, y, width, height, title) {
	// set up the GL window's properties
	mode(FL_RGB | FL_ALPHA | FL_DEPTH | FL_DOUBLE);

	// initialize track (radius = 10)
	track = new Track(10.0);

	timeElapsed = 0.0;
	timePerFrame = 0.01;

	// world view by default
	trainViewOn = false;

	// default camera
	cameraRotation = 0.0;
	cameraPitch = 40.0;
}


// destructor
P2GlWindow::~P2GlWindow() {
	// free up memory
	delete track;
}


// initialize GL operations (stuff that only need to be set once)
void P2GlWindow::initGL() {
	// which color we should fill in the color buffer when glClear is called
	glClearColor(.0, .0, .0, 1.);
	// enable the depth test (you would get random results if not enabled)
	glEnable(GL_DEPTH_TEST);
}


// overwrite the window draw function
void P2GlWindow::draw() {

	// initialize OpenGL if this is the first call to draw
	static bool firstTime = true; // Flag to control whether this is first time
	if (firstTime) { // The first time, we need to set up the necessary GL parameters
		initGL();
		firstTime = false;
	} // if

	// clear the color and depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);      

	///////////////////////// Set up camera... ////////////////////////

	// set up the projection transformation
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double aspectRatio = (double)w()/h();
	// parameters: field of view, aspect ratio, near plane, far plane
	gluPerspective(45., aspectRatio, .1, 100.);

	// set up the model view transformation
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	if (trainViewOn) {
		////////////////////// TRAIN VIEW /////////////////////////

		// find the position of the train
		double x, y, z;
		track->positionAtTime(timeElapsed, x, y, z);
		// Our eyes are actually a bit higher than the train...
		// We do this because we can't see the track clearly if our eyes
		// are at the same level as the track.
		z += 0.5;

		// find the direction vector of the train
		double dx, dy, dz;
		track->directionAtTime(timeElapsed, dx, dy, dz);

		// look at x+dx,y+dy,z+dz from x,y,z. 0,0,1 is the "up" vector.
		gluLookAt(x, y, z, x+dx, y+dy, z+dz, 0., 0., 1.);

	} else {
		///////////////////// WORLD VIEW //////////////////////////

		// first, look at the origin from (-30, 0, 0)
		gluLookAt(-30., 0., 0., 0., 0., 0., 0., 0., 1.);
		// bring the camera up about the y-axis
		glRotated(cameraPitch, 0., -1., 0.);
		// rotate it about the z-axis
		glRotated(cameraRotation, 0., 0., 1.);
	}

	/////////////////////// Draw stuff ////////////////////////

	// draw the ground plane, the track and the train
	drawGroundPlane();
	// draw the track
	track->draw();
	// the train itself blocks the view in train view, so we don't draw it
	if (!trainViewOn)
		drawTrain();
}

// advance animation by one frame, and draw
void P2GlWindow::advanceAnimation() {
	// if you are not sure what these variables mean, see P2GlWindow.h
	timeElapsed += timePerFrame;
	// one cycle = 1 unit of time
	if (timeElapsed > 1.0)
		timeElapsed -= 1.0;

	// "damage" the window - this tells Fltk something on the
	// screen is outdated, so it will redraw
	damage(1);
}

// toggle between train view and world view
void P2GlWindow::toggleView() {
	trainViewOn = !trainViewOn;
	// redraw
	damage(1);
}

// override event handler of window
int P2GlWindow::handle(int e) {
	switch(e) {
		case FL_ENTER:
		case FL_FOCUS:
		case FL_UNFOCUS:
			return 1;
		case FL_KEYDOWN: // keypress event
		{
			int c = Fl::event_key(); // the key pressed
			switch(c) {
				case FL_Right: // rotate world counterclockwise
					cameraRotation += 10.0;
					break;
				case FL_Left: // rotate world clockwise
					cameraRotation -= 10.0;
					break;
				case FL_Down: // rotate camera down
					if (cameraPitch < 85.0)
						cameraPitch += 10.0;
					break;
				case FL_Up: // rotate camera up
					if (cameraPitch > 25.0)
						cameraPitch -= 10.0;
					break;
				default:	// any other key? pass along to the parent
					return Fl_Gl_Window::handle(e);
			}
			damage(1); // redraw
			return 1;
		}
		default:	// we only handle keypresses - send everything else to the parent
			return Fl_Gl_Window::handle(e);
	}
}


// draw the ground plane
void P2GlWindow::drawGroundPlane() {
	// the ground plane is a 20x20 checkerboard
	// each square has size 2x2

	glBegin(GL_QUADS);

	for (int i = -10; i < 10; ++i)
		for (int j = -10; j < 10; ++j) {
			// set color (dark grey or light grey)
			if ((i+j)%2 == 0)
				glColor3d(.3, .3, .3);
			else
				glColor3d(.7, .7, .7);

			// the vertices of the square
			glVertex3d(i * 2. + 2, j * 2. + 2, 0.);
			glVertex3d(i * 2.    , j * 2. + 2, 0.);
			glVertex3d(i * 2.    , j * 2.    , 0.);
			glVertex3d(i * 2. + 2, j * 2.    , 0.);
		}

	glEnd();
}

// draw the train.
// note that this train is NOT correctly oriented. But in your project,
// you are required to correctly orient your train.
void P2GlWindow::drawTrain() {
	// find the position of the train
	double x, y, z;
	track->positionAtTime(timeElapsed, x, y, z);
	
	//------- draw the train (a small cube) -------//

	// translate to the train's position
	glPushMatrix();
	glTranslated(x, y, z);

	// draw the cube..
	const double size = 0.5;
	glBegin(GL_QUADS);

	// red
	glColor3d(1., 0., 0.);

	// face 1
	glVertex3d(size, -size, size);
	glVertex3d(-size, -size, size);
	glVertex3d(-size, -size, -size);
	glVertex3d(size, -size, -size);

	// face 2
	glVertex3d(size, size, size);
	glVertex3d(size, size, -size);
	glVertex3d(-size, size, -size);
	glVertex3d(-size, size, size);

	// face 3
	glVertex3d(size, -size, -size);
	glVertex3d(-size, -size, -size);
	glVertex3d(-size, size, -size);
	glVertex3d(size, size, -size);

	// face 4
	glVertex3d(size, size, size);
	glVertex3d(-size, size, size);
	glVertex3d(-size, -size, size);
	glVertex3d(size, -size, size);
	
	// face 5
	glVertex3d(size, -size, size);
	glVertex3d(size, -size, -size);
	glVertex3d(size, size, -size);
	glVertex3d(size, size, size);

	// face 6
	glVertex3d(-size, size, size);
	glVertex3d(-size, size, -size);
	glVertex3d(-size, -size, -size);
	glVertex3d(-size, -size, size);
	
	glEnd();

	// undo translation
	glPopMatrix();
}
